@@ -1,15 +1,17 @@
-/* Shared library add-on to iptables to add simple non load-balancing SNAT support. */
-#include <stdbool.h>
 #include <stdio.h>
-#include <netdb.h>
 #include <string.h>
 #include <stdlib.h>
-#include <getopt.h>
 #include <xtables.h>
 #include <net/netfilter/nf_nat.h>
-/* For 64bit kernel / 32bit userspace */
 #include <linux/netfilter_ipv4/ipt_SAME.h>
 
+enum {
+	O_TO_ADDR = 0,
+	O_NODST,
+	O_RANDOM,
+	F_RANDOM = 1 << O_RANDOM,
+};
+
 static void SAME_help(void)
 {
 	printf(
@@ -25,31 +27,23 @@ static void SAME_help(void)
 "				Randomize source port\n");
 }
 
-static const struct option SAME_opts[] = {
-	{.name = "to",     .has_arg = true,  .val = '1'},
-	{.name = "nodst",  .has_arg = false, .val = '2'},
-	{.name = "random", .has_arg = false, .val = '3'},
-	XT_GETOPT_TABLEEND,
+static const struct xt_option_entry SAME_opts[] = {
+	{.name = "to", .id = O_TO_ADDR, .type = XTTYPE_STRING,
+	 .flags = XTOPT_MAND},
+	{.name = "nodst", .id = O_NODST, .type = XTTYPE_NONE},
+	{.name = "random", .id = O_RANDOM, .type = XTTYPE_NONE},
+	XTOPT_TABLEEND,
 };
 
-static void SAME_init(struct xt_entry_target *t)
-{
-	struct ipt_same_info *mr = (struct ipt_same_info *)t->data;
-
-	/* Set default to 0 */
-	mr->rangesize = 0;
-	mr->info = 0;
-	mr->ipnum = 0;
-	
-}
-
 /* Parses range of IPs */
-static void
-parse_to(char *arg, struct nf_nat_range *range)
+static void parse_to(const char *orig_arg, struct nf_nat_range *range)
 {
-	char *dash;
+	char *dash, *arg;
 	const struct in_addr *ip;
 
+	arg = strdup(orig_arg);
+	if (arg == NULL)
+		xtables_error(RESOURCE_PROBLEM, "strdup");
 	range->flags |= IP_NAT_RANGE_MAP_IPS;
 	dash = strchr(arg, '-');
 
@@ -73,66 +67,37 @@ parse_to(char *arg, struct nf_nat_range *range)
 		if (range->min_ip > range->max_ip)
 			xtables_error(PARAMETER_PROBLEM, "Bad IP range \"%s-%s\"\n",
 				   arg, dash+1);
+	free(arg);
 }
 
-#define IPT_SAME_OPT_TO			0x01
-#define IPT_SAME_OPT_NODST		0x02
-#define IPT_SAME_OPT_RANDOM		0x04
-
-static int SAME_parse(int c, char **argv, int invert, unsigned int *flags,
-                      const void *entry, struct xt_entry_target **target)
+static void SAME_parse(struct xt_option_call *cb)
 {
-	struct ipt_same_info *mr
-		= (struct ipt_same_info *)(*target)->data;
+	struct ipt_same_info *mr = cb->data;
 	unsigned int count;
 
-	switch (c) {
-	case '1':
+	xtables_option_parse(cb);
+	switch (cb->entry->id) {
+	case O_TO_ADDR:
 		if (mr->rangesize == IPT_SAME_MAX_RANGE)
 			xtables_error(PARAMETER_PROBLEM,
 				   "Too many ranges specified, maximum "
 				   "is %i ranges.\n",
 				   IPT_SAME_MAX_RANGE);
-		if (xtables_check_inverse(optarg, &invert, NULL, 0, argv))
-			xtables_error(PARAMETER_PROBLEM,
-				   "Unexpected `!' after --to");
-
-		parse_to(optarg, &mr->range[mr->rangesize]);
+		parse_to(cb->arg, &mr->range[mr->rangesize]);
 		/* WTF do we need this for? */
-		if (*flags & IPT_SAME_OPT_RANDOM)
+		if (cb->xflags & F_RANDOM)
 			mr->range[mr->rangesize].flags 
 				|= IP_NAT_RANGE_PROTO_RANDOM;
 		mr->rangesize++;
-		*flags |= IPT_SAME_OPT_TO;
 		break;
-		
-	case '2':
-		if (*flags & IPT_SAME_OPT_NODST)
-			xtables_error(PARAMETER_PROBLEM,
-				   "Can't specify --nodst twice");
-		
+	case O_NODST:
 		mr->info |= IPT_SAME_NODST;
-		*flags |= IPT_SAME_OPT_NODST;
 		break;
-
-	case '3':	
-		*flags |= IPT_SAME_OPT_RANDOM;
+	case O_RANDOM:
 		for (count=0; count < mr->rangesize; count++)
 			mr->range[count].flags |= IP_NAT_RANGE_PROTO_RANDOM;
 		break;
-
-	default:
-		return 0;
 	}
-	
-	return 1;
-}
-
-static void SAME_check(unsigned int flags)
-{
-	if (!(flags & IPT_SAME_OPT_TO))
-		xtables_error(PARAMETER_PROBLEM,
-			   "SAME needs --to");
 }
 
 static void SAME_print(const void *ip, const struct xt_entry_target *target,
@@ -142,8 +107,8 @@ static void SAME_print(const void *ip, const struct xt_entry_target *target,
 	const struct ipt_same_info *mr = (const void *)target->data;
 	int random_selection = 0;
 	
-	printf("same:");
-	
+	printf(" same:");
+
 	for (count = 0; count < mr->rangesize; count++) {
 		const struct nf_nat_range *r = &mr->range[count];
 		struct in_addr a;
@@ -153,19 +118,17 @@ static void SAME_print(const void *ip, const struct xt_entry_target *target,
 		printf("%s", xtables_ipaddr_to_numeric(&a));
 		a.s_addr = r->max_ip;
 		
-		if (r->min_ip == r->max_ip)
-			printf(" ");
-		else
-			printf("-%s ", xtables_ipaddr_to_numeric(&a));
+		if (r->min_ip != r->max_ip)
+			printf("-%s", xtables_ipaddr_to_numeric(&a));
 		if (r->flags & IP_NAT_RANGE_PROTO_RANDOM) 
 			random_selection = 1;
 	}
 	
 	if (mr->info & IPT_SAME_NODST)
-		printf("nodst ");
+		printf(" nodst");
 
 	if (random_selection)
-		printf("random ");
+		printf(" random");
 }
 
 static void SAME_save(const void *ip, const struct xt_entry_target *target)
@@ -179,22 +142,20 @@ static void SAME_save(const void *ip, const struct xt_entry_target *target)
 		struct in_addr a;
 
 		a.s_addr = r->min_ip;
-		printf("--to %s", xtables_ipaddr_to_numeric(&a));
+		printf(" --to %s", xtables_ipaddr_to_numeric(&a));
 		a.s_addr = r->max_ip;
 
-		if (r->min_ip == r->max_ip)
-			printf(" ");
-		else
-			printf("-%s ", xtables_ipaddr_to_numeric(&a));
+		if (r->min_ip != r->max_ip)
+			printf("-%s", xtables_ipaddr_to_numeric(&a));
 		if (r->flags & IP_NAT_RANGE_PROTO_RANDOM) 
 			random_selection = 1;
 	}
 	
 	if (mr->info & IPT_SAME_NODST)
-		printf("--nodst ");
+		printf(" --nodst");
 
 	if (random_selection)
-		printf("--random ");
+		printf(" --random");
 }
 
 static struct xtables_target same_tg_reg = {
@@ -204,12 +165,10 @@ static struct xtables_target same_tg_reg = {
 	.size		= XT_ALIGN(sizeof(struct ipt_same_info)),
 	.userspacesize	= XT_ALIGN(sizeof(struct ipt_same_info)),
 	.help		= SAME_help,
-	.init		= SAME_init,
-	.parse		= SAME_parse,
-	.final_check	= SAME_check,
+	.x6_parse	= SAME_parse,
 	.print		= SAME_print,
 	.save		= SAME_save,
-	.extra_opts	= SAME_opts,
+	.x6_options	= SAME_opts,
 };
 
 void _init(void)
